section \<open>Tseitin Transformation (2 persons)\<close>

text \<open>
Since most SAT solvers insist on formulas in conjunctive normal form (CNF) as input,
but in general the CNF of a given formula may be exponentially larger, there is interest
in efficient transformations that produce a small equisatisfiable CNF for a given formula.
Probably the earliest and most well-known of these transformation is due to Tseitin.

In this project you will implement a two-step transformation of propositional formulas
into equisatisfiable CNFs and formally prove results about the complexity and that
the resulting CNFs are indeed equisatisfiable to the original formula.
\<close>

theory Project_Tseitin_Fresh
  imports Main
begin

subsection \<open>Syntax and Semantics\<close>

text \<open>
For the purposes of this project propositional formulas (with atoms of an arbitrary type)
are restricted to the following (functionally complete) connectives:
\<close>
datatype 'a form =
    Bot \<comment> \<open>the "always false" formula\<close>
  | Top \<comment> \<open>the "always true" formula\<close>
  | Var 'a \<comment> \<open>propositional variables\<close>
  | Neg "'a form" \<comment> \<open>negation\<close>
  | Disj "'a form" "'a form" \<comment> \<open>disjunction\<close>
  | Conj "'a form" "'a form" \<comment> \<open>conjunction\<close>

text \<open>
Define a function \<open>eval\<close> that evaluates the truth value of a formula with respect
to a given truth assignment \<open>\<alpha> :: 'a \<Rightarrow>bool\<close>.
\<close>
fun eval :: "('a \<Rightarrow> bool) \<Rightarrow> 'a form \<Rightarrow> bool"
  where
    "eval \<alpha> \<phi> = undefined"

text \<open>
Define a predicate \<open>sat\<close> that captures satisfiable formulas.
\<close>
definition sat :: "'a form \<Rightarrow> bool"
  where
    "sat \<phi> \<longleftrightarrow> undefined"


subsection \<open>Conjunctive Normal Forms\<close>

text \<open>
Literals are positive or negative variables.
\<close>
datatype 'a literal = P 'a | N 'a

text \<open>
A clause is a disjunction of literals, represented as a list of literals.
\<close>
type_synonym 'a clause = "'a literal list"

text \<open>
A CNF is a conjunction of clauses, represented as list of clauses.
\<close>
type_synonym 'a cnf = "'a clause list"

text \<open>
Implement a function \<open>of_cnf\<close> that, given a CNF (of @{typ "'a cnf"}, computes
a logically equivalent formula (of @{typ "'a form"}).
\<close>

fun of_cnf :: "'a cnf \<Rightarrow> 'a form"
  where
    "of_cnf cs = undefined"

subsection \<open>Tseitin Transformation\<close>

text \<open>
The idea of Tseitin's transformation is to assign to each subformula \<open>\<phi>\<close>
a label \<open>a\<^sub>\<phi>\<close> and use the following definitions

\<^item> \<open>a\<^sub>\<bottom> \<longleftrightarrow> \<bottom>\<close>
\<^item> \<open>a\<^sub>\<top> \<longleftrightarrow> \<top>\<close>
\<^item> \<open>a\<^sub>\<not>\<^sub>\<phi> \<longleftrightarrow> \<not> \<phi>\<close>
\<^item> \<open>a\<^sub>\<phi>\<^sub>\<or>\<^sub>\<psi> \<longleftrightarrow> (\<phi> \<or> \<psi>)\<close>
\<^item> \<open>a\<^sub>\<phi>\<^sub>\<and>\<^sub>\<psi> \<longleftrightarrow> (\<phi> \<and> \<psi>)\<close>

to recursively compute clauses \<open>tseitin \<phi>\<close> such that \<open>a\<^sub>\<phi> \<and> tseitin \<phi>\<close> and \<open>\<phi>\<close>
are equisatisfiable (that is, the former is satisfiable iff the latter is).

Define a function \<open>tseitin\<close> that computes the clauses corresponding to the above idea.
\<close>

fun tseitin :: "'a form \<Rightarrow> ('a form) cnf"
  where
    "tseitin \<phi> = undefined"

text \<open>
Prove that \<open>a\<^sub>\<phi> \<and> tseitin \<phi>\<close> are equisatisfiable.
\<close>

lemma tseitin_equisat:
  "sat (of_cnf ([P \<phi>] # tseitin \<phi>)) \<longleftrightarrow> sat \<phi>"
  sorry

text \<open>
Prove linear bounds on the number of clauses and literals by suitably
replacing \<open>n\<close> and \<open>num_literals\<close> below:
\<close>
lemma tseitin_num_clauses:
  "length (tseitin \<phi>) \<le> n * size \<phi>"
  sorry

lemma tseitin_num_literals:
  "num_literals (tseitin \<phi>) \<le> n * size \<phi>"
  sorry


subsection \<open>Fresh Variables\<close>

text \<open>
One of the problems in the tseitin transformation above is that the type of 
propositional variables is changed from @{typ 'a} to @{typ "'a form"}.

Define a function to rename variables in a CNF.
\<close>

fun rename_cnf :: "('a \<Rightarrow> 'b) \<Rightarrow> 'a cnf \<Rightarrow> 'b cnf"
  where
    "rename_cnf f cs = undefined"

text \<open>Think of a property such that renaming preserves satisfiability. Note that 
  injectivity is already defined in Isabelle (@{term inj} or @{const inj_on}.)\<close>

lemma "property f cs \<Longrightarrow> sat (of_cnf (rename_cnf f cs)) \<longleftrightarrow> sat (of_cnf cs)" sorry

text \<open>Next, we define a tseitin transformation which does not change the type of propositional variables.\<close>

definition tseitin_fresh :: "'your_type form \<Rightarrow> 'your_type cnf" where
  "tseitin_fresh \<phi> = (let 
      cs = [P \<phi>] # tseitin \<phi>;
      renaming = undefined
    in rename_cnf renaming cs)" 

text \<open>Implement a corresponding renaming function such that the following soundness property can be
  proved. Here, you also need to change the type-variable @{typ 'your_type}, where for this project
  it is perfectly fine to use a concrete type which has infinitely many elements, e.g., 
  @{typ nat} or @{typ int} or @{typ string}.\<close>

lemma tseitin_fresh: "sat \<phi> \<longleftrightarrow> sat (of_cnf (tseitin_fresh \<phi>))" sorry

text \<open>Your function definitions should be executable.\<close>

definition X :: 'your_type where "X = undefined" 
definition Y :: 'your_type where "Y = undefined" 
definition Z :: 'your_type where "Z = undefined" 

definition test_form :: "'your_type form" where
  "test_form = Neg (Conj (Disj (Neg (Var X)) (Var Z)) (Neg (Var Y)))" 

text \<open>The Isabelle command \<open>value (code) "tseitin_fresh test_form"\<close> should succeed.\<close>

end
